function [index,D] = adaptiveNeighbourhood(Y,center,opts,prnt)

% function [index,D] = adaptiveNeighbourhood(Y,center,opts,prnt)
%
% Chooses according to 'opts.thresh' or 'opts.Nsize' the neighbourhood
% structure of the actual subband. The patches of this band have already 
% been extracted in 'denoise_band' and are stored in 'Y'.  If 'prnt'>0 the 
% parent coefficient will be also included. There are two possible methods
% 1. 'opts.thresh' > 0 ,pick all j s.t. H_cj > (1-opts.thresh)*max_i H_ci
% 2. 'opts.thresh' < 0 ,pick j until sum_1^j H_cj > sum_i H_ci
%
% where 
%   'Y'         are all extracted neighbourhoods
%   'center'    is the index of the center coefficient
%   'opts'      some options, see 'newOptions'
%   'prnt'      1 if parent coefficient is included, 0 otherwise
%
%   'index'     indices within 'opts.bigblock' of included coeffs
%   'D'         size of the resulting neighborhood
%
% Peter-Vincent Gehler 16.March 2005

verbose = 0;

% ... compute the statistics ...
N = size(Y,2);
ycenter = Y(center,:);
y2 = Y.^2;
E22 = y2(center,:)*y2'/N;
E2q = (ycenter*Y'/N).^2;
Y2 = mean(y2')';
Eqq = Y2*Y2';

% ... and the cross cumulant.
C = E22 - 2*E2q - Eqq(center,:);

% ... sort according to its values ...
[B,index] = sort(C);

% Now remove the center pixel ...
b = B(index==center);
B(index==center) = [];
index(index==center) = [];

% ... and put it on last (first) position so it will surely be included...
index = [index,center];
rating = [B,b];

% ... now depending on the type of threshhold ...
if isfield(opts,'thresh') & opts.thresh < 0
    % 1. a threshold selection
    n = 0; D = 0;
    while n < abs(opts.thresh) * sum(rating(rating>0))
        D = D+1;
        % do not include negative cumulants
        if D>1 & rating(end-D) < 0
            D = D-1;
            break
        end
        n = sum(rating(end:-1:end-D));
    end
elseif isfield(opts,'thresh') & opts.thresh > 0
    marg = (1-opts.thresh) * max(rating(:));
    D = max(1,sum(rating(:)>=marg));
else
    % 2. simply the top opts.Nsize coefficients
    if ~isfield(opts,'Nsize')
        error('opts.Nsize is not specified (number of coefficients in neighbourhood)');
    end
    D = opts.Nsize;
end

% ... possibly include the parent if exsitent ...
if prnt
    index = [index(:,end:-1:end-D+1),prod(opts.bigblock)+1];
else
    index = index(:,end:-1:end-D+1);
end

% ... the size of the neighbourhood is ...
D = size(index,2);

% ... that's it. Now plot some stuff if interested.
if verbose > 1
    Ng = reshape(C,opts.bigblock(1),opts.bigblock(2));

    figure(1)
    myimagesc(Ng,1);
    axis off
    title('cross cumulant');

    T = zeros(opts.bigblock);
    T=T(:);
    T(index(1:end-1)) = 1;
    myimagesc(reshape(T,opts.bigblock(1),opts.bigblock(2)),2);
    axis off
    title('selected neighbourhood');
end
